# TODO: handlers for filters
# TODO: postordinate parser fails to get named
# TODO: step-to-parser command
# TODO: "current parser" on GUI is the one on top of the stack, while the argument of h_do_parse is not yet accounted for
# TODO: possible enhancement, caching the vtable type in Parser objects (allows searching by type)
# TODO: The parameter given to h_do_parse is not shown in the backtrace, which is confusing


# List of names of functions with H_RULEs declared, used by BreakpointManager
H_RULE_FUNCTIONS = [ "init_runlengthdecode_parser", "init_LZW_parser" ]

class Parser:
	def __init__(self, name, address):
		self.name = name
		self.address = address
		self.bytes_used = {}
	
	def name_parser(self, name):
		self.name = name

	# TODO: remove
	def get_name_or_placeholder(self):
		if self.name is None:
			return "Wait for it... (if you're reading this, you found a bug)"
		else:
			return self.name

	def add_mem_use(self, state, size):
		if self.bytes_used.setdefault(state, None) is None:
			self.bytes_used[state] = size
		else:
			self.bytes_used[state] += size

	def get_mem_use(self, state=None):
		if state is None:
			return self.bytes_used
		else:
			return self.bytes_used.setdefault(state, 0)

	def __str__(self):
		return "(" + str(self.name) + ", " + hex(self.address) + ")"

	# Return the highest per-arena allocation count
	# TODO: disambiguate "parse state" and "arena", possibly tracking both
	def get_arenamax(self):
		res = 0
		if self.bytes_used:
			res = max(self.bytes_used.values())
		return res

	def get_arenasum(self):
		res = 0
		if self.bytes_used:
			res = sum(self.bytes_used.values())
		return res

class ParserStack:
	def __init__(self, parse_state, arena):
		self.parse_state = parse_state
		self.arena = arena
		self.p_stack = []
		self.unclaimed_mem_use = 0

	def push(self, parser):
		self.p_stack.append(parser)

	def pop(self):
		return self.p_stack.pop()

	def peek(self):
		try:
			retval = self.p_stack[-1]
		except IndexError:
			retval = None
		return retval

	def set_state(self, state):
		self.parse_state = state
	# Shortcut for setting the name property of the parser on the top of stack
	# In terms of tracing, *most* calls to a parser look something like this with the packrat backend:
	# h_do_parse()
	#	parse_foo()
	#		perform_lowlevel_parse()
	
	# perform_lowlevel_parse() is called when the memo table at that position is not filled in yet.
	# it calls the corresponding parse_* virtual function via the vtable, but other than that does not have type information
	# it's probably possible to extract type information, by comparing vtable addresses, but that seems painful
	
	# parse_foo() is the parser's corresponding virtual function in the frontend, which does not have the equivalent of a "this" pointer
	
	# So what we do to keep track of parsers is incrementally filling in the details for both
	
	# h_do_parse() is the backend's "actually run the parser" function, but does not get called for some parsers
	# (apparently mostly it's for higher-order parsers)
	# also contains the decision logic about whether to call perform_lowlevel_parse()
	
	# possible scenarios:
	# h_do_parse()
	#	perform_lowlevel_parse()
	#		parse_foo()
	
	# h_do_parse()
	#	perform_lowlevel_parse()

	# h_do_parse()
	def name_top_parser(self, name):
		self.p_stack[-1].name_parser(name)

	def add_mem_use_each(self, size):
		for p in self.p_stack:
			p.bytes_used += size

	def add_mem_use_top(self, size):
		self.p_stack[-1].bytes_used += size
		
	def show_stack(self):
		print("stack would be printed here. Depth:", len(self.p_stack))
		#print([(p.get_name_or_placeholder(), hex(p.address)) for p in self.p_stack])

	def depth(self):
		return len(self.p_stack)

# Class that is responsible for bookkeeping throughout the entire parse
# NB, this is slightly different terminology than the hammer API implicitly uses:
# There, a parse is started by h_parse(), and it is associated with a parse state.
# This corresponds to the ParserStack above. TopLevelParse keeps track of all these.
# Subsequent h_do_parse()s with the same parser state are considered to belong to the same parse

print(": Initializing TopLevelParse")
top_level_parse = TopLevelParse()
# Approach 1: load the application, set breakpoints, execute stack commands on breakpoint hit, continue

class InitParserBreakpoint(gdb.Breakpoint):
	def stop(self):
		frame = gdb.selected_frame()
		block = frame.block()
		top_level_parse.init_parser()
		
		# This will also catch locals that aren't parsers, but it's not a problem in practice,
		# since h_parse() will never be called on them
		# If it becomes a problem after all, gdb.parse_and_eval() might be used to filter them out
		for p in block:
			top_level_parse.parser_objs[int(p.value(frame))] = Parser(p.name, int(p.value(frame)))

		return False

print(": Initializing BreakpointManager")
breakpoint_manager = BreakpointManager(H_RULE_FUNCTIONS)

class PDFMainBreakpoint(gdb.Breakpoint):
	def stop(self):
		breakpoints = breakpoint_manager.set_hammer_retq_breakpoints()
		#breakpoint_manager.hammer_retq_breakpoints = breakpoints

		return True

print(": Registering exit handler")
# Clean up by-address breakpoints in hammer when inferior exits.
# Caveat: Assumes there's a single inferior, the debugged parser, so no checking is done
# TODO: where to store breakpoints? TopLevelParse? A BreakpointManager class?i
def exit_handler(event):
	#breakpoints = [ perform_lowlevel_parse_ret, h_packrat_parse_ret ]
	#del_hammer_retq_breakpoints(breakpoints)
	breakpoint_manager.del_hammer_retq_breakpoints()

gdb.events.exited.connect(exit_handler)

print(": Setting PDFMain breakpoint")
# Break on main so that libhammer.so gets to load
main = PDFMainBreakpoint("main")




#init_parser = InitParserBreakpoint("pdf.c:1223")
#h_do_parse = HDoParseBreakpoint("h_do_parse")
#h_packrat_parse = HPackratParseBreakpoint("h_packrat_parse")
#perform_lowlevel_parse = PerformLowLevelParseBreakpoint("perform_lowlevel_parse")
#h_arena_malloc_raw = HArenaMallocRawBreakpoint("h_arena_malloc_raw")
print(": Setting Hammer library breakpoints")
print(":: Normal breakpoints")
breakpoint_manager.set_hammer_breakpoints()

# TODO: investigate GDB frame filters for rendering backtraces

#parse_action = ParserVirtualBreakpoint("parse_action")
#parse_choice = ParserVirtualBreakpoint("parse_choice")
#parse_sequence = ParserVirtualBreakpoint("parse_sequence")
#parse_difference = ParserVirtualBreakpoint("parse_difference")
#parse_many = ParserVirtualBreakpoint("parse_many")
#parse_and = ParserVirtualBreakpoint("parse_and")
#parse_attr_bool = ParserVirtualBreakpoint("parse_attr_bool")
#parse_bind = ParserVirtualBreakpoint("parse_bind")
#parse_bits = ParserVirtualBreakpoint("parse_bits")
#parse_butnot = ParserVirtualBreakpoint("parse_butnot")
#parse_charset = ParserVirtualBreakpoint("parse_charset")
#parse_ch = ParserVirtualBreakpoint("parse_ch")
#parse_end = ParserVirtualBreakpoint("parse_end")
#parse_endianness = ParserVirtualBreakpoint("parse_endianness")
#parse_epsilon = ParserVirtualBreakpoint("parse_epsilon")
#parse_ignore = ParserVirtualBreakpoint("parse_ignore")
#parse_ignoreseq = ParserVirtualBreakpoint("parse_ignoreseq")
#parse_indirect = ParserVirtualBreakpoint("parse_indirect")
#parse_int_range = ParserVirtualBreakpoint("parse_int_range")
#parse_not = ParserVirtualBreakpoint("parse_not")
#parse_nothing = ParserVirtualBreakpoint("parse_nothing")
#parse_optional = ParserVirtualBreakpoint("parse_optional")
#parse_permutation = ParserVirtualBreakpoint("parse_permutation")
#parse_skip = ParserVirtualBreakpoint("parse_skip")
#parse_seek = ParserVirtualBreakpoint("parse_seek")
#parse_tell = ParserVirtualBreakpoint("parse_tell")
#parse_token = ParserVirtualBreakpoint("parse_token")
#parse_unimplemented = ParserVirtualBreakpoint("parse_unimplemented")
#parse_put = ParserVirtualBreakpoint("parse_put")
#parse_get = ParserVirtualBreakpoint("parse_get")
#parse_whitespace = ParserVirtualBreakpoint("parse_whitespace")
#parse_xor = ParserVirtualBreakpoint("parse_xor")
#TODO: there would be less complaining about pending breakpoints if vtable breakpoints were set after running to main
print(":: Parser vtable breakpoints")
breakpoint_manager.set_parser_virtual_breakpoints()

# Commandline:
# $ gdb -ex "source /path/to/parser-name-instrumentation-gdb.py" --args /path/to/pdf /path/to/input.pdf

# run until main
print(": Running until main")
gdb.execute("run")

#plp_retq = locate_perform_lowlevel_parse_retq()
#perform_lowlevel_parse_ret = PerformLowLevelParseRetBreakpoint("*" + hex(plp_retq))
#hpp_retq = locate_h_packrat_parse_retq()
#h_packrat_parse_ret = HPackratParseRetBreakpoint("*" + hex(hpp_retq))

#i_p_retq = breakpoint_manager.locate_retq("init_parser")
#init_parser = InitParserBreakpoint("*" + hex(i_p_retq))
print(": Setting application breakpoints")
print(":: init_parser breakpoint")
breakpoint_manager.set_init_parser_breakpoint()

#hammer_retq_breakpoints = [perform_lowlevel_parse_ret, h_packrat_parse_ret]

#h_rule_breakpoints = {}

#for func in H_RULE_FUNCTIONS:
#	func_retq = locate_retq(func[0], func[1])
#	h_rule_breakpoints[func] = HRuleBreakpoint("*" + hex(func_retq))
print(":: RET breakpoints in functions with H_RULES")
breakpoint_manager.set_h_rule_breakpoints()


# TODO: the RET breakpoints in hammer break when "run" is executed again. figure out a way to automatically replace these

# Run until stop position, if set. Finish parsing otherwise
print(": Continuing execution")
gdb.execute("continue")

print(": Printing memory statistics")
print([(p.name, hex(p.address), p.bytes_used) for p in top_level_parse.parser_objs.values()])

# Approach 2: capture process trace with gdb, load the trace, execute stack commands on breakpoint hit, etc
