import unittest

class ParserCreation(unittest.TestCase):
	def setUp(self):
		self.parser_address = 12345
		self.parser_name = "foo"

	def test_parser_creation_with_address(self):
		test_parser = Parser(None, self.parser_address)
		self.assertIsInstance(test_parser.address, int)
		self.assertEqual(test_parser.address, 12345)
		self.assertIsNone(test_parser.name)

	def test_parser_creation_with_name(self):
		test_parser = Parser(self.parser_name, None)
		self.assertIsInstance(test_parser.name, str)
		self.assertEqual(test_parser.name, "foo")
		self.assertIsNone(test_parser.address)

	def test_parser_creation_with_name_and_address(self):
		test_parser = Parser(self.parser_name, self.parser_address)
		self.assertIsInstance(test_parser.name, str)
		self.assertEqual(test_parser.name, "foo")
		self.assertIsInstance(test_parser.address, int)
		self.assertEqual(test_parser.address, 12345)

class ParserMemUse(unittest.TestCase):
	def setUp(self):
		self.parser = Parser("foo", 12345)

	def test_add_mem_use(self):
		self.parser.add_mem_use(1, 50)
		self.parser.add_mem_use(1, 60)
		self.assertEqual(self.parser.get_mem_use(1), 110)

	def test_get_mem_use(self):
		self.parser.add_mem_use(1, 50)
		self.parser.add_mem_use(2, 60)
		self.assertEqual(self.parser.get_mem_use(1), 50)
		self.assertEqual(self.parser.get_mem_use(2), 60)
		self.assertEqual(self.parser.get_mem_use(), {1: 50, 2: 60})

	def test_arenamax(self):
		self.parser.add_mem_use(1, 50)
		self.parser.add_mem_use(2, 60)
		self.assertEqual(self.parser.get_arenamax(), 60)

	def test_arenasum(self):
		self.parser.add_mem_use(1, 50)
		self.parser.add_mem_use(2, 60)
		self.assertEqual(self.parser.get_arenasum(), 110)

class ParserNaming(unittest.TestCase):
	def setUp(self):
		self.parser = Parser(None, 12345)

	def test_name_parser_method(self):
		self.parser.name_parser("foo")
		self.assertEqual(self.parser.name, "foo")

	# This is enforced by TopLevelParse instead
	#@unittest.expectedFailure
	#def test_parser_cannot_be_named_twice(self):
	#	self.parser.name_parser("foo")
	#	self.parser.name_parser("bar")
	#	self.assertEqual(self.parser.name, "bar")

	#def test_first_parser_name_stays(self):
	#	self.parser.name_parser("foo")
	#	self.parser.name_parser("bar")
	#	self.assertEqual(self.parser.name, "foo")

class ParserStringRepresentation(unittest.TestCase):
	def setUp(self):
		self.parser = Parser("foo", 12345)

	def test_parser_string_rep(self):
		self.assertEqual(str(self.parser), "(foo, 0x3039)")

#unittest.main(exit=False)
