import unittest

class TopLevelParseStateManagement(unittest.TestCase):
	def setUp(self):
		self.top_level_parse = TopLevelParse()
		self.ps1 = ParserStack(16, 40)
		self.parser1 = Parser("foo", 32)

	def test_enter_h_packrat_parse(self):
		self.top_level_parse.enter_h_packrat_parse(32)
		self.assertIsInstance(self.top_level_parse.parser_stacks[0], ParserStack)

	def test_enter_h_do_parse_new_parser(self):
		ps1 = ParserStack(16, 40)
		self.top_level_parse.enter_h_packrat_parse(32)
		#simulating effects of enter_h_packrat_parse
		self.top_level_parse.enter_h_do_parse(16, 40, 32)
		parser1 = self.top_level_parse.parser_objs[32]
		self.assertIs(self.top_level_parse.h_do_parse_parser, self.top_level_parse.parser_objs[32])
		self.assertIsNone(parser1.name)
		self.assertEqual(parser1.address, 32)

	def test_enter_h_do_parse_known_parser(self):
		self.top_level_parse.enter_h_packrat_parse(32)
		parser1 = Parser("foo", 32)
		self.top_level_parse.parser_objs[32] = parser1 # Add the parser into the db by hand
		self.top_level_parse.enter_h_do_parse(16, 40, 32)
		self.assertIs(self.top_level_parse.h_do_parse_parser, parser1)
		self.assertIs(self.top_level_parse.h_do_parse_parser, self.top_level_parse.parser_objs[32])
		self.assertEqual(self.top_level_parse.h_do_parse_parser.name, "foo")
		self.assertEqual(self.top_level_parse.h_do_parse_parser.address, 32)

	def test_ehd_calls_first_h_do_parse_after_packrat_parse(self):
		parser1 = Parser("foo", 32)
		self.top_level_parse.enter_h_packrat_parse(parser1.address)
		self.top_level_parse.enter_h_do_parse(16, 40, parser1.address)
		ps1 = self.top_level_parse.peek_parserstack()
		self.assertEqual(ps1.parse_state, 16)

	def test_first_h_do_parse_after_packrat_parse(self):
		ps1 = ParserStack(None, 40)
		self.top_level_parse.parser_stacks.append(ps1)
		self.top_level_parse.first_h_do_parse_after_packrat_parse(16, 40)
		self.assertEqual(ps1.parse_state, 16)

	def test_return_from_h_packrat_parse(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.return_from_h_packrat_parse()
		self.assertEqual(self.top_level_parse.parser_stacks, [])

	def test_enter_perform_lowlevel_parse_known_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.top_level_parse.enter_perform_lowlevel_parse(32)
		self.assertIs(self.top_level_parse.parser_stacks[-1].p_stack[-1], self.parser1)
	def test_enter_perform_lowlevel_parse_new_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.enter_perform_lowlevel_parse(32)
		parser1 = self.top_level_parse.parser_stacks[-1].p_stack[-1]
		self.assertIsNone(parser1.name)
		self.assertEqual(parser1.address, 32)

	def test_return_from_perform_lowlevel_parse(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.top_level_parse.parser_stacks[-1].p_stack.append(self.parser1)
		self.top_level_parse.return_from_perform_lowlevel_parse()
		self.assertEqual(self.top_level_parse.parser_stacks[-1].p_stack, [])

	def test_enter_h_arena_malloc_raw(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.parser_stacks[-1].p_stack[-1].bytes_used[16], 256)

	def test_enter_h_arena_malloc_raw_no_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.parser_stacks[-1].unclaimed_mem_use, 256)

	def test_enter_h_arena_malloc_raw_no_stack(self):
		self.top_level_parse.enter_h_arena_malloc_raw(256)
		self.assertEqual(self.top_level_parse.unclaimed_mem_use, 256)

	def test_parse_virtual(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		parser2 = Parser(None, 32)
		self.top_level_parse.parser_stacks[-1].push(parser2)
		self.top_level_parse.parse_virtual("foo")
		self.assertEqual(parser2.name, "foo")

	def test_peek_parserstack(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		ps1 = self.top_level_parse.peek_parserstack()
		self.assertIs(ps1, self.ps1)

	def test_peek_parser(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser1 = self.top_level_parse.peek_parser()
		self.assertIs(parser1, self.parser1)

	def test_parser_by_address(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser1 = self.top_level_parse.parser_by_address(32)

	def test_parsers_by_name(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser2 = Parser("foo", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2)
		parserlist = self.top_level_parse.parsers_by_name("foo")
		self.assertEqual(parserlist, [self.parser1, parser2])

	def test_add_or_get_parser_add(self):
		parser2 = self.top_level_parse.add_or_get_parser(32)
		self.assertIsNone(parser2.name)
		self.assertEqual(parser2.address, 32)
		self.assertIs(self.top_level_parse.parser_objs[32], parser2)

	def test_add_or_get_parser_get(self):
		self.top_level_parse.parser_objs[32] = self.parser1
		parser1 = self.top_level_parse.add_or_get_parser(32)
		self.assertIs(parser1, self.parser1)

	def test_get_parser_top_per_arena_mem(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2)
		self.parser1.add_mem_use(16, 50)
		parser2.add_mem_use(16, 60)
		self.assertIs(self.top_level_parse.get_parser_top_per_arena_mem(), parser2)

	def test_get_parser_top_total_arena_mem(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)
		self.assertIs(self.top_level_parse.get_parser_top_total_arena_mem(), self.parser1)

	def test_get_avg_mem_use_per_arena(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)
		self.assertEqual(self.top_level_parse.get_avg_mem_use_per_arena(), {16: 55, 128: 50})

	def test_dump_memory_stats(self):
		self.top_level_parse.parser_stacks.append(self.ps1)
		self.top_level_parse.parser_objs[32] = self.parser1
		self.ps1.push(self.parser1)
		parser2 = Parser("bar", 48)
		self.top_level_parse.parser_objs[48] = parser2
		self.ps1.push(parser2)
		self.parser1.add_mem_use(16, 50)
		self.parser1.add_mem_use(128, 50)
		parser2.add_mem_use(16, 60)

		stats = self.top_level_parse.dump_memory_stats()
		self.assertEqual(stats, [('foo', '0x20', {16: 50, 128: 50}), ('bar', '0x30', {16: 60})])
